#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <iostream>
#include <string>
#include <map>
#include <regex>

#include "errhandling.h"
#include "socklib.h"

#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

/* Mini serveur WEB, LIFASR5, Nicolas Louvet, avril 2019*/

#define HTTP_ROOT "./web"       // racine du site web hébergé
#define HTTP_DFLT "/index.html" // nom de la page par défaut

using namespace std;

void print_usage(const string &name) {
  cerr << "Usage : " << name << " port" << endl;
}

void logmsg(string msg) {
  cerr << "LOG: " + msg << endl << flush;
}

// ==============================================================================================
// Déclaration des fonctions pour lire sur le disque et envoyer des vecteurs d'octets
// ==============================================================================================

// Fonction pour lire un fichier régulier dont le chemin est donné par l'argument path
// dans le système de fichier, et le stocker entièrement dans le vecteur t de type
// vector<char>, qu'il faut regarder comme un vecteur d'octets. L'appel retourne true en
// cas de succès (tout les octets du fichier ont été chargés), false en cas d'échec (qui
// peut avoir différentes origines, comme par exemple le fait que le fichier n'existe pas).
bool read_file(vector<char> &t, const string &path);

// Fonction pour envoyer tout les octets contenus dans un vecteur v de type
// vector<char> sur une socket s. L'appel retourne true en cas de succès
// (tout a été envoyé), false en cas d'échec (tout n'a pas pu être envoyé).
bool send_vec(int s, const vector<char> &v);

// ==============================================================================================
// Déclaration des fonctions pour recevoir les requêtes de la part d'un client
// ==============================================================================================

// Type pour recevoir une requête de la part d'un client.
typedef map<string, string> reqmsg;

// Permet de recevoir une requête HTTP/1.1 sur la socket de dialogue s.
// Retourne true en cas de succés, false en cas d'échec, et dans ce cas
// des messages d'erreur sont affichés.
bool reqmsg_recv(reqmsg &msg, int s);

// Cette fonction retourne sous la forme d'une string le type d'une requête reçue
// depuis un client : il s'agit du premier mot sur la ligne "reqline" dans la table
// de type reqmsg passée en argument. Dans notre cas, cela sera typiquement "GET",
// mais cela pour être aussi "POST", "HEAD", "PUT"... En cas d'échec, la fonction
// retourne la chaîne vide "".
string reqmsg_type(const reqmsg &msg);

// Si l'on sait que msg est une requête du type "GET", cette fonction permet d'obtenir
// l'url u de la page dont la requête fait l'objet, ainsi que la version v du protocole
// HTTP demandé par le client (les deux avec le type string). La fonction retourne true
// en cas de succès, false en cas d'échec.
bool reqmsg_split_reqget(const reqmsg &msg, string &u, string &v);

// ============================================================================================
// Réponses possibles à la requête d'un client
// ============================================================================================

// Juste une réponse 200 possible pour tester une version préliminaire du serveur.
void resp_test(int s) {
  const string crlf = "\r\n";
  string r;

  r += "<html>";
  r += "  <head>";
  r += "    <meta charset=\"UTF-8\">";
  r += "    <title>Page exemple</title>";
  r += "  </head>";
  r += "  <body>";
  r += "  <h1>Ceci est un exemple de page web :D</h1>";
  r += "  <p>Il va falloir faire mieux par la suite !<p>";
  r += "  </body>";
  r += "</html>";

  send_str(s, "HTTP/1.1 200 OK" + crlf);
  send_str(s, "Content-Type: text/html" + crlf);
  send_str(s, "Content-length: " + to_string(r.size()) + crlf);
  send_str(s, "Connection: close" + crlf);
  send_str(s, crlf);
  send_str(s, r);
}

// Réponse HTTP "simple" : envoie une réponse HTTP/1.1, avec sur la ligne de status le
// code fourni par le paramètre de même nom, et comme "raison" le message msg. Dans le
// corps de la réponse, une petite page HTML contenant le message msg est envoyée.
// Pour mémoire, code peut être :
// 404 -> NOT FOUND (le fichier demandé n'a pas été trouvé)
// 501 -> NOT IMPLEMENTED (le serveur ne supporte pas la requête effectuée)
// ...
void resp_code(int s, int code, const string &msg) {
  const string crlf("\r\n");
  string r;
  
  r += "<html>";
  r += "  <head>";
  r += "    <meta charset=\"UTF-8\">";
  r += "    <title>" + to_string(code) + ", " + msg +"</title>";
  r += "  </head>";
  r += "  <body>";
  r += "  <h1>" + to_string(code) + ", " + msg +"</h1>";
  r += "  </body>";
  r += "</html>";

  send_str(s, "HTTP/1.1 " + to_string(code) + " " + msg + crlf);
  send_str(s, "Content-Type: text/html" + crlf);
  send_str(s, "Content-length: " + to_string(r.size()) + crlf);
  send_str(s, "Connection: close" + crlf);
  send_str(s, crlf);
  send_str(s, r);
}

// Déduit un type MIME en fonction du chemin path passé en argument.
// Ce type MIME sera utilisé pour instuire le champ "Content-Type"
// dans une réponse HTTP à un client. La méthode est rudimentaire,
// mais elle devrait permettre de s'en sortir dans ce petit TP.
string content_type(const string &path) {
  smatch mch;
  string ext, ct;

  if(!regex_search(path, mch, regex(".*\\.([a-zAZ]+)"))) {
    cerr << "ERREUR, reqmsg::type : erreur lors du découpage." << endl << flush;
    ext = "";
  }
  else ext = mch[1];

  if(ext == "txt") ct = "text/plain";
  else if(ext == "htm" || ext == "html") ct = "text/html";
  else if(ext == "jpg" || ext == "jpeg") ct = "image/jpeg";
  else if(ext == "gif") ct = "image/gif";
  else ct = "application/octet-stream";

  return ct;
}

// Réponse 200 (OK, la requête a été accomplie) avec envoi du fichier.
bool resp_file(int s, const string &u) {
  // TODO2 : compléter le corps de la fonction.
  vector<char> t;

  if(!read_file(t, u)) {
    logmsg("resp_file : impossible d'ouvrir le fichier");
    return false;
  }

  const string crlf("\r\n");
  string r;
  send_str(s, "HTTP/1.1 200 OK" + crlf);
  send_str(s, "Content-Type: " + content_type(u) + crlf);
  send_str(s, "Content-length: " + to_string(t.size()) + crlf);
  send_str(s, "Connection: close" + crlf);
  send_str(s, crlf);

  send_vec(s, t);
  return true;
}

// ============================================================================================
// Gestion des signaux
// ============================================================================================

// Passe à 1 quand le serveur doit se terminer.
volatile sig_atomic_t quit = 0;

// Gestionnaire de signaux.
// Ce gestionnaire permettra au processus père de prendre
// en compte la mort de ses fils (réception de SIGCHLD), et
// éviter l'accumulation de processus zombies. On ré-installera
// le gestionnaire par défaut dans les fils (ça ne peut pas nuire).
// SIGINT permettra de mettre fin au serveur, en faisant passer
// la variable globale quit à 1. Les fils ignoreront ce signal.
void handler(int sig) {
// TODO4 : compléter la fonction des gestions pour les signaux SIGCHLD et SIGINT
}

// ============================================================================================
// Programme principal
// ============================================================================================


int main(int argc, char *argv[]) {
  if(argc < 2) {
    print_usage(argv[0]);
    return(1);
  }
  char *port = argv[1];

  // README1 : installation d'un gestionnaire pour le signal SIGCHLD.
  struct sigaction oldsa_chld, newsa_chld;
  sigaction(SIGCHLD, NULL, &newsa_chld);
  newsa_chld.sa_flags = SA_RESTART | SA_NOCLDSTOP;
  newsa_chld.sa_handler = handler;
  if(sigaction(SIGCHLD, &newsa_chld, &oldsa_chld) == -1)
    exit_error("installation du gestionnaire pour SIGCHLD");

  // README2 : installation d'un gestionnaire pour le signal SIGINT pour le père.
  struct sigaction partsa_int, chldsa_int;
  sigaction(SIGINT, NULL, &partsa_int);
  partsa_int.sa_handler = handler;
  sigaction(SIGINT, NULL, &chldsa_int);
  chldsa_int.sa_handler = SIG_IGN;
  if(sigaction(SIGINT, &partsa_int, NULL) == -1)
    exit_error("installation du gestionnaire pour SIGINT");

  int s;  // Socket d'écoute.
  int sd; // Socket de dialogue.
  string http_root(HTTP_ROOT); // Racine du site web hébergé.
  string http_dflt(HTTP_DFLT); // Page web par défaut.

  // Création de la socket d'écoute s.
  s = create_server_socket(port);
  if(s == -1) exit_error("lors de la création de la socket d'écoute");

  while(!quit) {

    sd = accept_connection(s);
    if(sd == -1) {
      if(quit) {
        logmsg("le serveur doit se terminer");
        break;
      }
      else exit_error("lors de l'attente d'une connexion");
    }

    cout << "Une connexion vient d'être acceptée." << endl;

    pid_t pid = fork();
    if(pid == -1)
      exit_error("lors de la création d'un fils");

    if(pid > 0) { // processus père
      logmsg("création du fils de PID " + to_string(pid) + " pour gérer un client");
    }
    else { // pid == 0, processus fils
      // On ferme la socket de connexion s.
      close(s);

      // TODO5 : remise en place du gestionnaire par défaut pour SIGCHLD.

      // TODO6 : les fils doivent ignorer le signal SIGINT.
      reqmsg req;

      if(!reqmsg_recv(req, sd)) {
        close(sd);
        exit_error("lors de la réception d'une requête");
      }

      if(reqmsg_type(req) == "GET") {
        string u, v;
        if(!reqmsg_split_reqget(req, u, v)) {
          close(sd);
          exit_error("lors de l'analyse d'une requête GET");
        }
        logmsg("url demandée : " + u);
        if(u == "/") u = http_dflt;
        if(!resp_file(sd, http_root + u)) resp_code(sd, 404, "Not Found");
      }
      else resp_code(sd, 501, "Not Implemented");

      close(sd);

      return EXIT_SUCCESS;
    }
  }

  logmsg("le serveur se termine");

  // TODO7 : attendre la fin des éventuels fils encore en cours d'exécution

  return EXIT_SUCCESS;
}

// ==============================================================================================
// Définition des fonctions pour lire sur le disque et envoyer des vecteurs d'octets
// ==============================================================================================

bool read_file(vector<char> &t, const string &path) {
  static const int buflen = 256;
  char buf[buflen];
  int nbrd;

  int fdin = open(path.data(), O_RDONLY);
  if(fdin == -1) {
    print_warn("read_file : impossible d'ouvrir le fichier " + path + " en lecture");
    return false;
  }

  t.clear();

  do {
    nbrd = read(fdin, buf, buflen);
    if(nbrd < 0) {
      if(nbrd < 0) cerr << "Erreur : " << strerror(errno) << endl;
      close(fdin);
      return 1;
    }
    for(int i = 0; i < nbrd; i++) t.push_back(buf[i]);
  } while(nbrd > 0);

  close(fdin);

  return true;
}

bool send_vec(int s, const vector<char> &v) {
  // TODO1 : compléter le corps de la fonction.
  int res = send_all(s, v.data(), v.size());
  if(res < 0) {
    print_error("send_vec");
    return false;
  }
  return true;
}

// ==============================================================================================
// Définition des fonctions pour recevoir les requêtes de la part d'un client
// ==============================================================================================

void strip_line(string &line) {
  if(!line.empty() && line.back() == '\n') line.pop_back();
  if(!line.empty() && line.back() == '\r') line.pop_back();
}

bool reqmsg_recv(reqmsg &msg, int s) {
  regex rgx("(.+)\\s*:\\s*(.+)");
  smatch mch;
  string line;

  msg.clear();

  if(recv_line(s, line, '\n') == -1) {
    cerr << "ERREUR, reqmsg::recv : erreur à la réception de la 1ère ligne." << endl << flush;
    return false;
  }
  strip_line(line);
  logmsg(line);
  if(line == "") {
    cerr << "ERREUR, reqmsg::recv : connexion fermée." << endl << flush;
    return false;
  }
  msg["reqline"] = line;

  while(true) {
    if(recv_line(s, line, '\n') == -1) {
      cerr << "ERREUR, reqmsg::recv : erreur lors de la réception d'une ligne." << endl << flush;
      return false;
    }
    strip_line(line);
    logmsg(line);
    if(line == "") break;
    if(!regex_search(line, mch, rgx)) {
      cerr << "ERREUR, reqmsg::recv : erreur lors du découpage." << endl << flush;
      return false;
    }
    msg[mch[1]] = mch[2];
  }

  return true;
}

string reqmsg_type(const reqmsg &msg) {
  smatch mch;
  if(!regex_search(msg.at("reqline"), mch, regex("([A-Z]+)\\s+.*"))) {
    logmsg("reqmsg_type, erreur lors du découpage");
    return "";
  }
  return mch[1];
}

bool reqmsg_split_reqget(const reqmsg &msg, string &u, string &v) {
  smatch mch;
  if(!regex_search(msg.at("reqline"), mch, regex("GET\\s+([A-Za-z0-9/._-]+)\\s+(HTTP/[0-9].[0-9])"))) {
    logmsg("reqmsg_split_reqget, erreur lors du découpage");
    return false;
  }
  u = mch[1];
  v = mch[2];
  return true;
}

